/*
Copyright (C) 2005 Observatoire de Paris-Meudon

Ce programme est un logiciel libre ; vous pouvez le redistribuer et/ou le modifier conformment aux dispositions de la Licence Publique Gnrale GNU, telle que publie par la Free Software Foundation ; version 2 de la licence, ou encore ( votre choix) toute version ultrieure.

Ce programme est distribu dans l'espoir qu'il sera utile, mais SANS AUCUNE GARANTIE ; sans mme la garantie implicite de COMMERCIALISATION ou D'ADAPTATION A UN OBJET PARTICULIER. Pour plus de dtail, voir la Licence Publique Gnrale GNU .

Vous devez avoir reu un exemplaire de la Licence Publique Gnrale GNU en mme temps que ce programme ; si ce n'est pas le cas, crivez  la Free Software Foundation Inc., 675 Mass Ave, Cambridge, MA 02139, Etats-Unis.
*/

package ciel3d;

import java.applet.Applet;
import java.awt.*;
import java.awt.event.*;
import java.net.*;
import java.util.Vector;

import xml.*;

/**
 * Carte du ciel en 3D
 *
 * Dimensions: 500x500
 *
 * Paramtres:
 * etoiles - nom du fichier XML avec les noms et coordonnes des toiles
 *           arbre :
 *           ETOILES
 *              +-- ETOILE @nom
 *                    +-- AD   sous forme dcimale en degrs ou HH MM SS.dd
 *                    +-- DEC  sous forme dcimale en degrs ou DD MM SS.dd
 *                    +-- DIST  (distance en parsec) ou    PLX  (parallaxe en mas)
 *                    +-- NP - nombre de plantes
 *                    +-- SPECT - type spectral      ou    B-V
 */
public class Ciel3D extends Applet {
    
    private static final int espace = 15;
    private static final String baseChemin = "systeme.php?etoile=";
    private int dimx=500, dimy=500;
    private String[] tnoms;
    private double[] ctx1; // quivalent (x,y,z) des coordonnes ad/dec
    private double[] cty1;
    private double[] ctz1;
    private double[] ctx2; // position modifie
    private double[] cty2;
    private double[] ctz2;
    private int[] tx; // position (x,y) sur la carte + profondeur z
    private int[] ty;
    private int[] tz;
    private boolean[] tvis; // point visible ?
    private double[] tdist; // distance en pc
    private int[] tnp; // nombre de plantes
    private Color[] tcoul; // couleurs
    private ImageCanvas ic;
    private Matrice tmpmat, amat;
    
    public void init() {
        setLayout(new BorderLayout());
        setBackground(new Color(0, 34, 62));
        setForeground(Color.white);
        String fichierXML = getParameter("etoiles");
        
        lectureXML(fichierXML);
        ic = new ImageCanvas(dimx, dimy, tnoms, tdist, tnp, tx, ty, tz, tvis, tcoul, this);
        add(ic, BorderLayout.CENTER);
        Label info = new Label(Messages.get("rotations_souris") + "     " + Messages.get("zoom_touches"));
        add(info, BorderLayout.SOUTH);
        addKeyListener(ic);
        ic.requestFocus();
    }
    
    protected void lectureXML(String fichierXML) {
        XMLTree tree = new XMLTree(null, "", null, "");
        URL hurle;
        try {
            if (fichierXML.startsWith("http://"))
                hurle = new URL(fichierXML);
            else
                hurle = new URL(getDocumentBase(), fichierXML);
        } catch (MalformedURLException ex) {
            System.err.println("MalformedURLException: " + ex.getMessage());
            return;
        }
        Parser_XML p = new Parser_XML(hurle, System.out, tree);
        p.parse(0);
        if (!"ETOILES".equalsIgnoreCase(tree.tag)) {
            System.err.println("Mauvais fichier XML !");
            return;
        }
        XMLTree etoile;
        String nom, sad, sdec, sdist, snp, spect, splx, sbmv;
        Vector vnoms = new Vector();
        Vector vad = new Vector();
        Vector vdec = new Vector();
        Vector vdist = new Vector();
        Vector vnp = new Vector();
        Vector vspect = new Vector();
        Vector vplx = new Vector();
        Vector vbmv = new Vector();
        
        for (etoile=tree.first_child; etoile!=null; etoile=etoile.next_brother) {
            nom = etoile.getAttVal("nom");
            sad = etoile.getNode("AD").val;
            sdec = etoile.getNode("DEC").val;
            XMLTree ndist = etoile.getNode("DIST");
            if (ndist != null)
                sdist = ndist.val;
            else
                sdist = null;
            XMLTree nnp = etoile.getNode("NP");
            if (nnp != null)
                snp = nnp.val;
            else
                snp = null;
            XMLTree nspect = etoile.getNode("SPECT");
            if (nspect != null)
                spect = nspect.val;
            else
                spect = null;
            XMLTree nplx = etoile.getNode("PLX");
            if (nplx != null)
                splx = nplx.val;
            else
                splx = null;
            XMLTree nbmv = etoile.getNode("B-V");
            if (nbmv != null)
                sbmv = nbmv.val;
            else
                sbmv = null;
            if (!"".equals(sdist)) {
                vnoms.addElement(nom);
                Double ad;
                if (sad.indexOf(' ') == -1)
                    ad = Double.valueOf(sad);
                else
                    ad = new Double(stringToCoord(sad)*360./24.);
                vad.addElement(ad);
                Double dec;
                if (sdec.indexOf(' ') == -1)
                    dec = Double.valueOf(sdec);
                else
                    dec = new Double(stringToCoord(sdec));
                vdec.addElement(dec);
                if (sdist != null)
                    vdist.addElement(sdist);
                if (snp != null)
                    vnp.addElement(snp);
                if (spect != null)
                    vspect.addElement(spect);
                if (splx != null)
                    vplx.addElement(splx);
                if (sbmv != null)
                    vbmv.addElement(sbmv);
            }
        }
        
        vnoms.addElement("Soleil");
        vad.addElement(new Double(0));
        vdec.addElement(new Double(0));
        if (vdist.size() == 0)
            vplx.addElement("1000"); // fausse valeur corrige plus tard
        else
            vdist.addElement("0");
        if (vnp.size() != 0)
            vnp.addElement("9");
        if (vspect.size() == 0)
            vbmv.addElement("0.62"); // approximatif, valeur exacte = ?
        else
            vspect.addElement("G2V");
        
        tnoms = new String[vnoms.size()];
        for (int i=0; i<vnoms.size(); i++)
            tnoms[i] = (String)vnoms.elementAt(i);
        if (vdist.size() == 0) {
            double[] tplx = versDoubles(vplx);
            tdist = new double[tplx.length];
            for (int i=0; i<tplx.length; i++)
                tdist[i] = Math.round(1000 * 1000. / tplx[i]) / 1000.;
            tdist[tdist.length-1] = 0;
        } else
            tdist = versDoubles(vdist);
        if (vnp.size() != 0)
            tnp = versEntiers(vnp);
        else
            tnp = null;
        if (vspect.size() == 0) {
            tcoul = new Color[vbmv.size()];
            for (int i=0; i<vbmv.size(); i++) {
                sbmv = (String)vbmv.elementAt(i);
                double bmv;
                try {
                    bmv = (new Double(sbmv)).doubleValue();
                } catch (NumberFormatException ex) {
                    System.err.println("NumberFormatException : " + ex.getMessage());
                    bmv = 0.62;
                }
                tcoul[i] = couleurEtoile(bmvVersTemperature(bmv));
            }
        } else {
            tcoul = new Color[vspect.size()];
            for (int i=0; i<vspect.size(); i++) {
                spect = (String)vspect.elementAt(i);
                tcoul[i] = spectreVersCouleur(spect);
            }
        }
        
        tmpmat = new Matrice();
        amat = new Matrice();
        if (tnp != null)
            tmpmat.zoom(2);
        else
            tmpmat.zoom(40);
        amat.multiplication(tmpmat);
        
        ctx1 = new double[vad.size()];
        cty1 = new double[vad.size()];
        ctz1 = new double[vad.size()];
        for (int i=0; i<vad.size(); i++) {
            double ad = ((Double)vad.elementAt(i)).doubleValue() * Math.PI/180;
            double dec = ((Double)vdec.elementAt(i)).doubleValue() * Math.PI/180;
            double dec2 = Math.PI/2 - dec;
            ctx1[i] = tdist[i] * Math.sin(dec2) * Math.cos(ad);
            cty1[i] = tdist[i] * Math.sin(dec2) * Math.sin(ad);
            ctz1[i] =  tdist[i] * Math.cos(dec2);
        }
        ctx2 = new double[vad.size()];
        cty2 = new double[vad.size()];
        ctz2 = new double[vad.size()];
        tx = new int[vad.size()];
        ty = new int[vad.size()];
        tz = new int[vad.size()];
        tvis = new boolean[vad.size()];
        calculCoord();
    }
    
    protected void calculCoord() {
        int cx = dimx/2;
        int cy = dimy/2;
        
        tmpmat.init();
        tmpmat.multiplication(amat);
        tmpmat.translation(cx, cy, 0);
        tmpmat.transformation(ctx1, cty1, ctz1, ctx2, cty2, ctz2);
        for (int i=0; i<ctx2.length; i++) {
            tx[i] = (int)Math.round(ctx2[i]);
            ty[i] = (int)Math.round(cty2[i]);
            tz[i] = (int)Math.round(ctz2[i]);
            tvis[i] = tx[i] > 0 && tx[i] < dimx && ty[i] > 0 && ty[i] < dimy;
        }
    }
    
    public void rotation(Point pt1, Point pt2) {
        tmpmat.init();
        double xtheta = (pt2.y - pt1.y) / 100.0;
        double ytheta = (pt2.x - pt1.x) / 100.0;
        tmpmat.rotationx(xtheta);
        tmpmat.rotationy(ytheta);
        amat.multiplication(tmpmat);
        calculCoord();
        ic.changerDonnees(tx, ty, tz, tvis);
    }
    
    public void rotation(String direction) {
        tmpmat.init();
        double xtheta = 0;
        double ytheta = 0;
        if ("haut".equals(direction))
            xtheta = -0.1;
        else if ("bas".equals(direction))
            xtheta = 0.1;
        else if ("gauche".equals(direction))
            ytheta = -0.1;
        else if ("droite".equals(direction))
            ytheta = 0.1;
        if (xtheta != 0)
            tmpmat.rotationx(xtheta);
        if (ytheta != 0)
            tmpmat.rotationy(ytheta);
        amat.multiplication(tmpmat);
        calculCoord();
        ic.changerDonnees(tx, ty, tz, tvis);
    }
    
    public void zoom(boolean plusgrand) {
        tmpmat.init();
        double v;
        if (plusgrand)
            v = 3./2.;
        else
            v = 2./3.;
        tmpmat.zoom(v);
        amat.multiplication(tmpmat);
        calculCoord();
        ic.changerDonnees(tx, ty, tz, tvis);
    }
    
    public void ouvrirEtoile(String nom) {
        try {
            URL hurle;
            if ("Soleil".equals(nom))
                hurle = new URL(getDocumentBase(), "systeme_solaire/solaire.html");
            else
                hurle = new URL(getDocumentBase(), baseChemin + URLEncoder.encode(nom));
            getAppletContext().showDocument(hurle, "_blank");
        } catch (MalformedURLException ex) {
            System.err.println("MalformedURLException: " + ex.getMessage());
        }
    }
    
    private double u0(double x) {
        // Heavyside
        if (x <= 0)
            return(0);
        else
            return(1);
    }
    
    private double sgn(double x) {
        if (x < 0)
            return(-1);
        else
            return(1);
    }
    
    protected static Color spectreVersCouleur(String spectre) {
        return(couleurEtoile(spectreVersTemperature(spectre)));
    }
    
    protected static Color couleurEtoile(double temperature) {
        // fausses couleurs, pour rendre le graphique plus lisible
        // (a permet de voir facilement le type spectral des toiles)
        int r,v,b;
        if (temperature < 3000) {
            r = 255;
            v = 0;
            b = 0;
        } else if (temperature < 6000) {
            r = 255;
            v = (int)(255*(temperature-3000)/3000);
            b = 0;
        } else if (temperature < 10000) {
            b = (int)(255*(temperature-6000)/4000);
            r = 255;
            v = 255;
        } else if (temperature < 50000) {
            b = 255;
            r = 255 - (int)(255*(temperature-10000)/40000);
            v = 255 - (int)(255*(temperature-10000)/40000);
        } else {
            r = 0;
            v = 0;
            b = 255;
        }
        return(new Color(r, v, b));
    }
    
    protected static double spectreVersTemperature(String spectre) {
        if ("pulsar".equals(spectre))
            return(1E6); // trs approximatif !
        
        // sur la squence principale
        // cf http://www.earth.uni.edu/~morgan/astro/course/Notes/section2/spectraltemps.html
        String[] tspec2 = {"O5", "O6", "O7", "O8", "O9", "B0", "B1", "B2", "B3", "B5", "B6", "B7",
            "B8", "B9", "A0", "A1", "A2", "A3", "A4", "A5", "A7", "F0", "F2", "F3", "F5", "F6",
            "F7", "F8", "G0", "G1", "G2", "G5", "G8", "K0", "K1", "K2", "K3", "K4", "K5", "K7",
            "M0", "M1", "M2", "M3", "M4", "M5", "M6", "M7", "M8"};
        double[] ttemp = {54000, 45000, 43300, 40600, 37800, 29200, 23000, 21000, 17600, 15200,
            14300, 13500,
            12300, 11400, 9600, 9330, 9040, 8750, 8480, 8310, 7920, 7350, 7050, 6850, 6700, 6550,
            6400, 6300, 6050, 5930, 5800, 5660, 5440, 5240, 5110, 4960, 4800, 4600, 4400, 4000,
            3750, 3700, 3600, 3500, 3400, 3200, 3100, 2900, 2700};
        if (spectre != null && spectre.length() > 1) {
            String spec2 = spectre.substring(0, 2);
            for (int i=0; i<tspec2.length; i++)
                if (tspec2[i].equals(spec2))
                    return(ttemp[i]);
        }
        String spec1;
        if (spectre != null && !"".equals(spectre))
            spec1 = spectre.substring(0, 1);
        else
            spec1 = null;
        if ("O".equals(spec1))
            return(35000);
        else if ("B".equals(spec1))
            return(13500);
        else if ("A".equals(spec1))
            return(8100);
        else if ("F".equals(spec1))
            return(6500);
        else if ("G".equals(spec1))
            return(5400);
        else if ("K".equals(spec1))
            return(4000);
        else if ("M".equals(spec1))
            return(2600);
        else {
            System.err.println("spectre non trouv dans la table: " + spectre);
            return(5400);
        }
    }
    
    protected static double bmvVersTemperature(double bmv) {
        // B-V  -> temprature
        double c1 = 3.979145;
        double c2 = -0.654499;
        double c3 = 1.74069;
        double c4 = -4.608815;
        double c5 = 6.7926;
        double c6 = -5.39691;
        double c7 = 2.19297;
        double c8 = -.359496;
        double logt = c1 + c2*bmv + c3*Math.pow(bmv,2) + c4*Math.pow(bmv,3) + c5*Math.pow(bmv,4) +
            c6*Math.pow(bmv,5) + c7*Math.pow(bmv,6) + c8*Math.pow(bmv,7);
        return(Math.pow(10, logt));
    }
    
    public static double stringToCoord(String s) throws NumberFormatException {
        int ind1, ind2;
        String s1,s2,s3;
        double d1,d2,d3;
        int signe;
        
        ind1 = s.indexOf(' ');
        if (ind1 == -1)
            throw new NumberFormatException();
        ind2 = s.substring(ind1 + 1).indexOf(' ');
        if (ind2 == -1)
            throw new NumberFormatException();
        ind2 += ind1 + 1;
        
        s1 = s.substring(0, ind1);
        s2 = s.substring(ind1 + 1, ind2);
        s3 = s.substring(ind2 + 1);
        
        d1 = (new Double(s1)).doubleValue();
        d2 = (new Double(s2)).doubleValue();
        d3 = (new Double(s3)).doubleValue();
        
        if (d1 < 0)
            signe = -1;
        else
            signe = 1;
        return(d1 + signe*d2/60 + signe*d3/3600);
    }
    
    // Vector de String -> tableau de double
    private static double[] versDoubles(Vector v) {
        if (v == null || v.size() == 0)
            return(null);
        double[] t = new double[v.size()];
        for (int i=0; i<v.size(); i++) {
            if ("".equals((String)v.elementAt(i)))
                t[i] = 0;
            else {
                try {
                    t[i] = (new Double((String)v.elementAt(i))).doubleValue();
                } catch (NumberFormatException ex) {
                    System.err.println(((String)v.elementAt(i)) + " -> NumberFormatException: " + ex.getMessage());
                    t[i] = 0;
                }
            }
        }
        return(t);
    }

    // Vector de String -> tableau de int
    private static int[] versEntiers(Vector v) {
        if (v == null || v.size() == 0)
            return(null);
        int[] t = new int[v.size()];
        for (int i=0; i<v.size(); i++) {
            try {
                t[i] = (new Integer((String)v.elementAt(i))).intValue();
            } catch (NumberFormatException ex) {
                System.err.println(((String)v.elementAt(i)) + " -> NumberFormatException: " + ex.getMessage());
                t[i] = 0;
            }
        }
        return(t);
    }
}
