/*
Copyright (C) 2005 Observatoire de Paris-Meudon

Ce programme est un logiciel libre ; vous pouvez le redistribuer et/ou le modifier conformment aux dispositions de la Licence Publique Gnrale GNU, telle que publie par la Free Software Foundation ; version 2 de la licence, ou encore ( votre choix) toute version ultrieure.

Ce programme est distribu dans l'espoir qu'il sera utile, mais SANS AUCUNE GARANTIE ; sans mme la garantie implicite de COMMERCIALISATION ou D'ADAPTATION A UN OBJET PARTICULIER. Pour plus de dtail, voir la Licence Publique Gnrale GNU .

Vous devez avoir reu un exemplaire de la Licence Publique Gnrale GNU en mme temps que ce programme ; si ce n'est pas le cas, crivez  la Free Software Foundation Inc., 675 Mass Ave, Cambridge, MA 02139, Etats-Unis.
*/

package ciel3d;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.lang.reflect.Method;

/**
 * Zone d'affichage de la carte
 */
public class ImageCanvas extends Canvas implements MouseListener, MouseMotionListener, KeyListener {
    private Dimension tailleImage;
    private int dimx, dimy;
    private int indEtoileSel;
    private String[] tnoms;
    private double[] tdist;
    private int[] tnp;
    private int[] tx;
    private int[] ty;
    private int[] tz;
    private boolean[] tvis;
    private Color[] tcoul;
    private int[] ordre;
    private Ciel3D app;
    private Image offImage = null;
    Point mousePos;
    
    public ImageCanvas(int dimx, int dimy, String[] tnoms, double[] tdist, int[] tnp,
            int[] tx, int[] ty, int[] tz, boolean[] tvis, Color[] tcoul, Ciel3D app) {
        this.dimx = dimx;
        this.dimy = dimy;
        this.tnoms = tnoms;
        this.tdist = tdist;
        this.tnp = tnp;
        this.tx = tx;
        this.ty = ty;
        this.tz = tz;
        this.tvis = tvis;
        this.tcoul = tcoul;
        this.app = app;
        tailleImage = new Dimension(dimx, dimy);
        offImage = app.createImage(dimx, dimy);
        indEtoileSel = -1;
        if (tx != null && ty != null) {
            addMouseListener(this);
            addMouseMotionListener(this);
            addKeyListener(this);
        }
        ordre = new int[tz.length];
        for (int i=0; i<tz.length; i++)
            ordre[i] = i;
        verifierOrdre();
    }
    
    public void changerDonnees(int[] tx, int[] ty, int[] tz, boolean[] tvis) {
        this.tx = tx;
        this.ty = ty;
        this.tz = tz;
        this.tvis = tvis;
        indEtoileSel = -1;
        verifierOrdre();
        repaint();
    }
    
    public Dimension getPreferredSize() {
        return tailleImage;
    }
    
    public Dimension getMinimumSize() {
        return tailleImage;
    }
    
    public void paint(Graphics g) {
        if (offImage != null) {
            Graphics offg = offImage.getGraphics();
            offg.setColor(Color.black);
            offg.fillRect(0, 0, dimx, dimy);
            dessinerEtoiles(offg);
            
            if (indEtoileSel != -1)
                dessinerSelection(offg);
            
            g.drawImage(offImage, 0, 0, this);
        }
    }
    
    protected void verifierOrdre() {
        int k;
        for (int i=0; i<tz.length; i++)
            for (int j=i+1; j<tz.length; j++) {
                if (tz[ordre[j]] < tz[ordre[i]]) {
                    k = ordre[i];
                    ordre[i] = ordre[j];
                    ordre[j] = k;
                }
            }
    }
    
    protected void dessinerEtoiles(Graphics g) {
        g.setColor(Color.yellow);
        for (int i=0; i<tx.length; i++) {
            int oi = ordre[i];
            if (tvis[oi]) {
                int r = (tz[oi] + 100) / 40;
                if (r < 1)
                    r = 1;
                //g.fillOval(tx[oi] - r, ty[oi] - r, 2*r, 2*r);
                dessinEtoile(g, tx[oi] - r, ty[oi] - r, 2*r, 2*r, tcoul[oi]);
            }
        }
        g.setColor(Color.black);
    }
    
    protected void dessinEtoile(Graphics g, int x, int y, int longueur, int hauteur, Color couleurEtoile) {
        int rayon = longueur/2;
        int cx = x + rayon;
        int cy = y + rayon;
        int rouge = couleurEtoile.getRed();
        int vert = couleurEtoile.getGreen();
        int bleu = couleurEtoile.getBlue();
        Color couleur;
        
        double ratio;
        for (int i=rayon; i>0; i-=2) {
            ratio = i*1.0/rayon;
            couleur = new Color(255 - (int)Math.round((255-rouge)*ratio),
                255 - (int)Math.round((255-vert)*ratio),
                255 - (int)Math.round((255-bleu)*ratio));
            g.setColor(couleur);
            g.fillOval(cx-i, cy-i, i*2, i*2);
        }
    }
    
    protected void dessinerSelection(Graphics g) {
        int xe = tx[indEtoileSel];
        int ye = ty[indEtoileSel];
        int r = (tz[indEtoileSel] + 100) / 40;
        if (r < 1)
            r = 1;
        g.setColor(Color.yellow);
        g.drawOval(xe - r, ye - r, 2 * r, 2 * r);
        g.setColor(Color.red);
        g.drawOval(xe - r - 1, ye - r - 1, 2 * r + 2, 2 * r + 2);
        String nom = tnoms[indEtoileSel];
        String dist = Messages.get("distance") + ": " + tdist[indEtoileSel] + " pc";
        String snp;
        if (tnp == null)
            snp = null;
        else if (tnp[indEtoileSel] == 0)
            snp = Messages.get("nb_planetes") + ": ?";
        else
            snp = Messages.get("nb_planetes") + ": " + tnp[indEtoileSel];
        int w = g.getFontMetrics().stringWidth(nom);
        int w2 = g.getFontMetrics().stringWidth(dist);
        int w3;
        if (snp != null)
            w3 = g.getFontMetrics().stringWidth(snp);
        else
            w3 = 0;
        if (w2 > w)
            w = w2;
        if (w3 > w)
            w = w3;
        if (xe + w + 6 > dimx)
            xe = dimx - w - 6;
        if (ye + 42 > dimy)
            ye = dimy - 42;
        g.setColor(new Color(255, 255, 203));
        int h;
        if (snp != null)
            h = 42;
        else
            h = 28;
        g.fillRoundRect(xe+2, ye+2, w+4, h, 6, 6);
        g.setColor(Color.black);
        g.drawString(nom, xe+4, ye+4+12);
        g.drawString(dist, xe+4, ye+4+24);
        if (snp != null)
            g.drawString(snp, xe+4, ye+4+36);
    }
    
    public void update(Graphics g) {
        paint(g);
    }
    
    protected int chercherEtoile(int xs, int ys) {
        double distmin = 1E10;
        int indice = -1;
        int xe, ye;
        int xemxs, yemys;
        double tempdistmin;
        for (int i=0; i<tnoms.length; i++)
            if (tvis[i]) {
                xe = tx[i];
                ye = ty[i];
                xemxs = xe-xs;
                yemys = ye-ys;
                tempdistmin = xemxs*xemxs + yemys*yemys;
                if (tempdistmin < distmin) {
                    distmin = tempdistmin;
                    indice = i;
                }
        }
        if (distmin < 20*20*2)
            return(indice);
        else
            return(-1);
    }
    
    public void mouseClicked(MouseEvent e) {
        if (tnp != null) {
            int indice = chercherEtoile(e.getX(), e.getY());
            if (indice != -1)
                app.ouvrirEtoile(tnoms[indice]);
        }
    }
    
    public void mousePressed(MouseEvent e) {
    }
    
    public void mouseReleased(MouseEvent e) {
        mousePos = null;
    }
    
    public void mouseEntered(MouseEvent e) {
    }
    
    public void mouseExited(MouseEvent e) {
    }
    
    public void mouseDragged(MouseEvent e) {
        Point nouvPt = e.getPoint();
        if (mousePos != null)
            app.rotation(mousePos, nouvPt);
        mousePos = nouvPt;
    }
    
    public void mouseMoved(MouseEvent e) {
        int indice = chercherEtoile(e.getX(), e.getY());
        if (indice != indEtoileSel) {
            if (indEtoileSel == -1)
                setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
            else if (indice == -1)
                setCursor(Cursor.getDefaultCursor());
            indEtoileSel = indice;
            repaint();
        }
    }
    
    public void keyTyped(KeyEvent e) {
        if (e.getKeyChar() == '+')
            app.zoom(true);
        else if (e.getKeyChar() == '-')
            app.zoom(false);
    }
    
    public void keyPressed(KeyEvent e) {
        if (e.getKeyCode() == KeyEvent.VK_UP)
            app.rotation("haut");
        else if (e.getKeyCode() == KeyEvent.VK_DOWN)
            app.rotation("bas");
        else if (e.getKeyCode() == KeyEvent.VK_LEFT)
            app.rotation("gauche");
        else if (e.getKeyCode() == KeyEvent.VK_RIGHT)
            app.rotation("droite");
    }
    
    public void keyReleased(KeyEvent e) {
    }
}
